# (c) Peter Kankowski, 2009 http://www.strchr.com kankowski@narod.ru
# Script for creating character properties table
# Requires Python 3.x, generates ANSI C code
import re, math

properties = [
	("lower", "a-z"),
	("upper", "A-Z"),
	("digit", "0-9"),
	("alnum", "a-zA-Z0-9"),
	("space", " \t\r\n\v\f"),
	("punct", "!-/:-@[-`{-~"),
	("xdigit", "0-9a-fA-F"),
	("graph", "!-~"),
]
# Instructions:
# * put the name of binary property (e.g., "vowel") into properties array.
# * list characters for which the property is true (e.g., "aeiouy").
#   You can use ranges (e.g., a-z or 0-9). To include literal dash ("-"),
#   make it the first or the last character of the string.

# The script generates several code snippets using different methods
# (see http://www.strchr.com/table-driven#character_properties).
# You can use any of the generated snippets. Enjoy!

INT_BITS   = 32      # The number of bits in int
SHORT_BITS = 16      # Ditto for short
CHAR_BITS  =  8      # and char
CHAR_VALUES = 2 ** CHAR_BITS
INDENT = '    '

# Expand "a-z" ranges
def expand_ranges(prop):
	def range_replacement(match):
		r = range(ord(match.group(1)), ord(match.group(2))+1)
		return ''.join([chr(i) for i in r])
	return re.sub(r'(.)-(.)', range_replacement, prop)

# Select the smallest data type to accommodate all properties of a single character
def select_data_type(nbits):
	if nbits > INT_BITS:
		raise Exception("too much properties")
	elif nbits > SHORT_BITS:
		return 'unsigned int'
	elif nbits > CHAR_BITS:
		return 'unsigned short'
	else:
		return 'unsigned char'

# Return a character in single quotes if it's printable, or a character code otherwise
def get_character(ch):
	return "'" + ch + "'" if ' ' <= ch <= '~' else ord(ch)

# Print 256-element table containing multiple properties
def print_multiprop_table(prop_names, props):
	ITEMS_PER_LINE = 16
	datatype = select_data_type(len(props))
	print('static const ' + datatype + ' properties[] = {', end='')
	
	for ch in range(CHAR_VALUES):
		val = 0
		for i, prop in enumerate(props):
			if prop.find(chr(ch)) != -1:
				val |= 1 << i

		if ch % ITEMS_PER_LINE == 0:
			print('\n' + INDENT, end='')
		print('{0:3},'.format(val), end='')
	print('\n};')
	
	for i, prop in enumerate(prop_names):
		print('#define IS_{0:15s} (properties[ch] & 0x{1:02x})'.format(prop.upper() + '(ch)', 1 << i))
	print('\n')

# Print 256-bit (32-byte) table containing the bits for one property
def print_single_prop_table(name, prop):
	print('static const unsigned int ' + name + '[] = {\n' + INDENT, end='')
	for i in range(CHAR_VALUES // INT_BITS):
		val = 0
		for bit in range(INT_BITS):
			ch = i * INT_BITS + bit
			if prop.find(chr(ch)) != -1:
				val |= 1 << bit
		
		print('0x{0:x}, '.format(val), end='')
	print('\n};')
	print('#define IS_{0}(ch) (({1}[(ch) >> {2}]) & (1 << ((ch) & {3})))\n'.format(
		name.upper(), name, int(math.log(INT_BITS, 2)), INT_BITS - 1))

# The difference between the largest and the smallest character of the set
def chars_range(prop):
	prop = list(prop)
	return ord(max(prop)) - ord(min(prop)) + 1

# Test if the range is continuous
def is_continuous_range(prop):
	prop = set(prop)
	return len(prop) == ( ord(max(prop)) - ord(min(prop)) + 1 )

# Use the single comparison method (http://strchr.com/checking_if_point_belongs_to_interval)
def print_continuous(name, prop):
	prop = list(prop)
	min_ch = min(prop)
	max_ch = max(prop)
	print("#define IS_{0}(ch) ((unsigned)(ch) - {1} < {2})\n".format(
		name.upper(), get_character(min_ch), ord(max_ch) - ord(min_ch) + 1))

# Print the bits for one property in one machine word
def print_single_word(name, prop):
	prop = list(prop)
	min_ch = ord(min(prop))
	# Avoid subtraction when possible
	if ord(max(prop)) < INT_BITS:
		min_ch = 0
	val = 0
	for ch in prop:
		val |= 1 << (ord(ch) - min_ch)
	
	print("inline int IS_{0}(int ch) {{\n"
			"{4}int x = ch - {2};\n"
			"{4}return (0x{1:x} >> x) & ((x & 0x{3:x}) == 0);\n"
			"}}\n".format(name.upper(), val, get_character(chr(min_ch)),
			2 ** INT_BITS - INT_BITS, INDENT))

# The main program

# Print multiple-properties table
prop_names = [prop[0] for prop in properties]
props = [expand_ranges(prop[1]) for prop in properties]
print_multiprop_table(prop_names, props)

# Select the best method for each property and print the bits
for prop in properties:
	prop_name = prop[0]
	prop = expand_ranges(prop[1])
	if is_continuous_range(prop):
		print_continuous(prop_name, prop)
	elif chars_range(prop) <= INT_BITS:
		print_single_word(prop_name, prop)
	else:
		print_single_prop_table(prop_name, prop)

